.\" Copyright (c) 1983 Regents of the University of California.
.\" All rights reserved.  The Berkeley software License Agreement
.\" specifies the terms and conditions for redistribution.
.\"
.\"	@(#)printcap.5	6.4 (Berkeley) 5/14/86
.\"
.TH PRINTCAP 5 "May 14, 1986"
.UC 5
.ad
.SH NAME
printcap \- printer capability data base
.SH SYNOPSIS
/etc/printcap
.SH DESCRIPTION
.I Printcap
is a simplified version of the
.IR termcap (5)
data base
used to describe line printers.  The spooling system accesses the
.I printcap
file every time it is used, allowing dynamic
addition and deletion of printers.  Each entry in the data base
is used to describe one printer.  This data base may not be
substituted for, as is possible for 
.IR termcap ,
because it may allow accounting to be bypassed.
.PP
The default printer is normally 
.IR lp ,
though the environment variable PRINTER
may be used to override this.  Each spooling utility supports an option,
.BI \-P printer,
to allow explicit naming of a destination printer.
.PP
Refer to the
.ul
4.3BSD Line Printer Spooler Manual
for a complete discussion on how setup the database for a given printer.
.SH CAPABILITIES
Refer to
.IR termcap (5)
for a description of the file layout.
.nf

.ta \w'k0-k9  'u +\w'Type  'u +\w'``/usr/spool/lpd\'\'  'u
\fBName	Type	Default	Description\fR
af	str	NULL	name of accounting file
br	num	none	if lp is a tty, set the baud rate (ioctl call)
cf	str	NULL	cifplot data filter
df	str	NULL	tex data filter (DVI format)
fc	num	0	if lp is a tty, clear flag bits (sgtty.h)
ff	str	``\ef''	string to send for a form feed
fo	bool	false	print a form feed when device is opened
fs	num	0	like `fc' but set bits
gf	str	NULL	graph data filter (plot (3X) format)
hl	bool	false	print the burst header page last
ic	bool	false	driver supports (non standard) ioctl to indent printout
if	str	NULL	name of text filter which does accounting
lf	str	``/dev/console''	error logging file name
lo	str	``lock''	name of lock file
lp	str	``/dev/lp''	device name to open for output
mx	num	1000	maximum file size (in BUFSIZ blocks), zero = unlimited
nd	str	NULL	next directory for list of queues (unimplemented)
nf	str	NULL	ditroff data filter (device independent troff)
of	str	NULL	name of output filtering program
pc	num	200	price per foot or page in hundredths of cents
pl	num	66	page length (in lines)
pw	num	132	page width (in characters)
px	num	0	page width in pixels (horizontal)
py	num	0	page length in pixels (vertical)
rf	str	NULL	filter for printing FORTRAN style text files
rg	str	NULL	restricted group. Only members of group allowed access
rm	str	NULL	machine name for remote printer
rp	str	``lp''	remote printer name argument
rs	bool	false	restrict remote users to those with local accounts
rw	bool	false	open the printer device for reading and writing
sb	bool	false	short banner (one line only)
sc	bool	false	suppress multiple copies
sd	str	``/usr/spool/lpd''	spool directory
sf	bool	false	suppress form feeds
sh	bool	false	suppress printing of burst page header
st	str	``status''	status file name
tf	str	NULL	troff data filter (cat phototypesetter)
tr	str	NULL	trailer string to print when queue empties
vf	str	NULL	raster image filter
xc	num	0	if lp is a tty, clear local mode bits (tty (4))
xs	num	0	like `xc' but set bits
.fi
.PP
If the local line printer driver supports indentation, the daemon
must understand how to invoke it.
.SH FILTERS
The
.IR lpd (8)
daemon creates a pipeline of
.I filters
to process files for various printer types.
The filters selected depend on the flags passed to
.IR lpr (1).
The pipeline set up is:
.RS
.PP
.nf
.ta 0.5i +1i
\-p	pr | if	regular text + \fIpr\fP(1)
none	if	regular text
\-c	cf	cifplot
\-d	df	DVI (tex)
\-g	gf	\fIplot\fP(3)
\-n	nf	ditroff
\-f	rf	Fortran
\-t	tf	troff
\-v	vf	raster image
.fi
.RE
.PP
The
.B if
filter is invoked with arguments:
.PP
	\fIif\fP [ \fB\-c\fP ] \fB\-w\fPwidth \fB\-l\fPlength \fB\-i\fPindent \fB\-n\fP login \fB\-h\fP host acct-file
.PP
The
.B \-c
flag is passed only if the
.B \-l
flag (pass control characters literally)
is specified to
.IR lpr .
.I Width
and
.I length
specify the page width and length
(from
.B pw
and
.B pl
respectively) in characters.
The
.B \-n
and
.B \-h
parameters specify the login name and host name of the owner
of the job respectively.
.I Acct-file
is passed from the
.B af
.I printcap
entry.
.PP
If no
.B if
is specified,
.B of
is used instead,
with the distinction that
.B of
is opened only once,
while
.B if
is opened for every individual job.
Thus,
.B if
is better suited to performing accounting.
The
.B of
is only given the
.I width
and
.I length
flags.
.PP
All other filters are called as:
.PP
	\fIfilter\fP \fB\-x\fPwidth \fB\-y\fPlength \fB\-n\fP login \fB\-h\fP host acct-file
.PP
where
.I width
and
.I length
are represented in pixels,
specified by the
.B px
and
.B py
entries respectively.
.PP
All filters take
.I stdin
as the file,
.I stdout
as the printer,
may log either to
.I stderr
or using
.IR syslog (3),
and must not ignore
.SM SIGINT.
.SH LOGGING
Error messages generated by the line printer programs themselves
(that is, the
.IR lp *
programs)
are logged by
.IR syslog (3)
using the
.I LPR
facility.
Messages printed on
.I stderr
of one of the filters
are sent to the corresponding
.B lf
file.
The filters may, of course, use
.I syslog
themselves.
.PP
Error messages sent to the console have a carriage return and a line
feed appended to them, rather than just a line feed.
.SH "SEE ALSO"
termcap(5),
lpc(8),
lpd(8),
pac(8),
lpr(1),
lpq(1),
lprm(1)
.br
.ul
4.3BSD Line Printer Spooler Manual
